#  Copyright (c) 2022 by Amplo.

import json

import numpy as np
import pandas as pd
import pytest
from sklearn.datasets import fetch_california_housing, load_iris

# TODO: Make use of the dummy data creator
#  from Amplo.Utils.testing import (DummyDataSampler, make_data, make_cat_data, make_num_data)  # noqa: E501
from amplo.automl import DataProcessor
from amplo.utils.data import check_dataframe_quality


@pytest.fixture(scope="class", params=["regression", "classification"])
def make_mode(request):
    mode = request.param
    target = "target"
    if mode == "classification":
        x, y = load_iris(return_X_y=True, as_frame=True)
    elif mode == "regression":
        x, y = fetch_california_housing(return_X_y=True, as_frame=True)
    else:
        raise NotImplementedError("Invalid mode")
    x.columns = [f"Feature_{i}" for i in range(len(x.columns))]
    request.cls.mode = mode
    request.cls.target = target
    request.cls.data = pd.concat([x, y.to_frame(target)], axis=1)
    yield


@pytest.mark.usefixtures("make_mode")
class TestMode:
    def test_mode(self):
        dp = DataProcessor(self.target)
        cleaned = dp.fit_transform(self.data)
        assert check_dataframe_quality(cleaned)


class TestDataProcessor:
    def test_interpolation(self):
        dp = DataProcessor()
        data = pd.DataFrame({"a": [1, np.nan, np.nan, np.nan, 5], "b": [1, 2, 3, 4, 5]})
        cleaned = dp.fit_transform(data)
        assert cleaned["a"].tolist() == [1, 0, 0, 0, 5]

    def test_type_detector(self):
        dp = DataProcessor()
        data = pd.DataFrame(
            {
                "a": ["a", "b", "c", "d", "a"],
                "b": [1, 2, 3, 4, 5],
                "c": [
                    "2020-01-01",
                    "2020-01-02",
                    "2020-01-03",
                    "2020-01-04",
                    "2020-01-05",
                ],
            }
        )
        cleaned = dp.fit_transform(data)
        assert pd.api.types.is_integer_dtype(cleaned["b"])
        assert pd.api.types.is_datetime64_any_dtype(cleaned["c"])
        assert "a_b" in cleaned
        assert "a_c" in cleaned
        assert "a_d" in cleaned

    def test_type_detector_with_nan(self):
        dp = DataProcessor()
        data = pd.DataFrame(
            {
                "a": ["a", "b", np.nan, "c", "d"],
                "b": [1, 2, 3, 4, np.nan],
                "c": ["2020-01-01", np.nan, "2020-01-03", "2020-01-04", "2020-01-05"],
            }
        )
        cleaned = dp.fit_transform(data)
        assert pd.api.types.is_float_dtype(cleaned["b"])
        assert pd.api.types.is_datetime64_any_dtype(cleaned["c"])
        assert "a_b" in cleaned
        assert "a_c" in cleaned
        assert "a_d" in cleaned

    def test_missing_values(self):
        data = pd.DataFrame(
            {
                "a": ["a", "b", np.nan, "c", "d"],
                "b": [1, 2, 3, 4, np.nan],
                "c": ["2020-01-01", np.nan, "2020-01-03", "2020-01-04", "2020-01-05"],
                "d": [1, 2, 3, 4, 5],
            }
        )

        # Remove rows
        dp = DataProcessor(missing_values="remove_rows")
        cleaned = dp.fit_transform(data)
        assert (
            len(cleaned) == 3
        ), cleaned.head()  # the nan in the categorical is omitted

        # Remove cols
        dp = DataProcessor(missing_values="remove_cols")
        cleaned = dp.fit_transform(data)
        assert (
            len(cleaned.keys()) == 6
        ), cleaned.head()  # Categorical NaN is allowed, so 5 categoricals and d

        # Replace with 0
        dp = DataProcessor(missing_values="zero")
        cleaned = dp.fit_transform(data)
        assert (cleaned.loc[2, ["a_b", "a_c", "a_d"]] == 0).all()
        assert cleaned["b"][4] == 0
        assert cleaned["c"][1] == 0

        # Interpolate
        dp = DataProcessor(missing_values="interpolate")
        cleaned = dp.fit_transform(data)
        assert cleaned.isna().sum().sum() == 0, cleaned.head()

        # Fill with mean
        dp = DataProcessor(missing_values="mean")
        cleaned = dp.fit_transform(data)
        assert (cleaned.loc[2, ["a_b", "a_c", "a_d"]] == 0).all()
        assert cleaned["b"][4] == 2.5
        assert cleaned["c"][1] == pd.to_datetime("2020-01-03 6:00:00", utc=True)

    def test_classification_target(self):
        data = pd.DataFrame(
            {
                "a": [2, 2, 1, 1, 2],
                "b": ["class1", "class2", "class1", "class2", "class1"],
            }
        )

        # Numerical not starting at 0
        dp = DataProcessor(target="a")
        cleaned = dp.fit_transform(data)
        assert set(cleaned["a"]) == {0, 1}

        # Categorical
        dp = DataProcessor(target="b")
        cleaned = dp.fit_transform(data)
        assert set(cleaned["b"]) == {0, 1}

    def test_outliers(self):
        x = pd.DataFrame(
            {
                "a": [*(23 * [1]), 1e15],
                "target": np.linspace(0, 1, 24).tolist(),
            }
        )

        # Clip
        dp = DataProcessor(outlier_removal="clip", target="target")
        xt = dp.fit_transform(x)
        assert xt.max().max() < 1e15, "Outlier not removed"
        assert not xt.isna().any().any(), "NaN found"

        # z-score
        dp = DataProcessor(outlier_removal="z-score", target="target")
        xt = dp.fit_transform(x)
        assert xt.max().max() < 1e15, "Outlier not removed"
        assert not xt.isna().any().any(), "NaN found"
        assert np.isclose(
            dp.transform(pd.DataFrame({"a": [1e14], "b": [1]})).max().max(), 1e14
        )
        assert dp.transform(pd.DataFrame({"a": [1e16], "b": [1]})).max().max() == 1

        # Quantiles
        dp = DataProcessor(outlier_removal="quantiles", target="target")
        xt = dp.fit_transform(x)
        assert xt.max().max() < 1e15, "Outlier not removed"
        assert not xt.isna().any().any(), "NaN found"
        assert dp.transform(pd.DataFrame({"a": [2], "b": [-2]})).max().max() == 0

    def test_duplicates(self):
        # Create a DataFrame that contains two columns with the same name
        x = pd.DataFrame({"a": [1, 2, 1], "a_copy": [1, 2, 1], "b": [3, 1, 3]})
        x = x.rename(columns={"a_copy": "a"})
        dp = DataProcessor()
        xt = dp.fit_transform(x)
        assert len(xt) == 2, "Didn't remove duplicate rows"
        assert len(xt.keys()) == 2, "Didn't remove duplicate columns"

    def test_constants(self):
        x = pd.DataFrame({"a": [1, 1, 1, 1, 1], "b": [1, 2, 3, 5, 6]})
        dp = DataProcessor()
        xt = dp.fit_transform(x)
        assert "a" not in xt.keys(), "Didn't remove constant column"

    def test_dummies(self):
        x = pd.DataFrame({"a": ["a", "b", "c", "b", "c", "a"]})
        dp = DataProcessor(cat_cols=["a"])
        xt = dp.fit_transform(x)
        assert "a" not in xt.keys(), "'a' still in keys"
        assert "a_b" in xt.keys(), "a_b missing"
        assert "a_c" in xt.keys(), "a_c missing"
        xt2 = dp.transform(pd.DataFrame({"a": ["a", "c"]}))
        assert np.allclose(
            xt2.values,
            pd.DataFrame({"a_a": [1, 0], "a_b": [0, 0], "a_c": [0, 1]}).values,
        ), "Converted not correct"

    def test_nan_categorical(self):
        # Setup
        df = pd.DataFrame({"a": ["hoi", np.nan, np.nan, np.nan]})
        dp = DataProcessor()
        df = dp.fit_transform(df)

        # Tests
        assert (
            "a" in dp.cat_cols
        ), "NaN categorical column not recognised as categorical."
        assert "a_hoi" in list(
            df.keys()
        ), f"Categorical column not properly converted: {df.keys()}"

    def test_settings(self):
        # todo add tests for different arguments, was failing for
        #  outlier_removal=z-score.
        target = "target"
        x = pd.DataFrame(
            {
                "a": ["a", "b", "c", "b", "c", "a"],
                "b": [1, 1, 1, 1, 1, 1],
                target: ["a", "b", "c", "b", "c", "a"],
            }
        )
        dp = DataProcessor(cat_cols=["a"], target=target)
        xt = dp.fit_transform(x)
        assert len(xt.drop(target, axis=1).keys()) == x["a"].nunique()
        settings = dp.get_settings()
        dp2 = DataProcessor()
        dp2.load_settings(settings)
        assert isinstance(dp2.get_settings()["_label_encodings"], list)
        xt2 = dp2.transform(pd.DataFrame({"a": ["a", "b"], "b": [1, 2]}))
        assert np.allclose(
            pd.DataFrame(
                {"b": [1.0, 2.0], "a_a": [1, 0], "a_b": [0, 1], "a_c": [0, 0]}
            ).values,
            xt2.values,
        )

    def test_pruner(self):
        x = pd.DataFrame({"a": ["a", "b", "c", "b", "c", "a"], "b": [1, 1, 1, 1, 1, 1]})
        dp = DataProcessor()
        dp.fit_transform(x)
        dp.prune_features(["b"])
        assert dp.int_cols == ["b"]
        assert dp.cat_cols == []

    def test_json_serializable(self):
        x = pd.DataFrame(
            {"a": ["a", "b", "c", "b", "c", "a"], "b": [1, "missing", 1, 1, 1, 1]}
        )
        for o in ["quantiles", "z-score", "clip", "none"]:
            for mv in ["remove_rows", "remove_cols", "interpolate", "mean", "zero"]:
                dp = DataProcessor(outlier_removal=o, missing_values=mv)
                dp.fit_transform(x)
                json.dumps(dp.get_settings())

    def test_cat_target(self):
        df = pd.DataFrame(
            {"a": ["a", "b", "c", "b", "c", "a"], "b": [1, 2, 3, 4, 5, 6]}
        )
        dp = DataProcessor(target="a")
        xt = dp.fit_transform(df)
        assert "a" in xt
        assert np.allclose(xt["a"].values, np.array([0, 1, 2, 1, 2, 0]))

    def test_target_encoding(self):
        target = "target"
        orig_labels = pd.Series(["a", "b", "c", "b", "c", "a"], name=target)
        dp = DataProcessor(target=target)
        enc_labels = pd.Series(dp.encode_labels(orig_labels))
        assert enc_labels.nunique() == orig_labels.nunique()
        assert enc_labels.min() == 0, "Encoding must start at zero"
        assert pd.api.types.is_integer_dtype(
            enc_labels
        ), "Encoding must be of dtype `int`"
        dec_labels = pd.Series(dp.decode_labels(enc_labels))
        assert (
            dec_labels == orig_labels
        ).all(), "Decoding does not result in original dataframe"



#  Copyright (c) 2022 by Amplo.

import catboost
import numpy as np
from sklearn.model_selection import train_test_split


class CatBoostClassifier:
    _estimator_type = "classifier"
    default_params = {
        "verbose": 0,
        "n_estimators": 1000,
        "allow_writing_files": False,
    }
    has_predict_proba = True

    def __init__(self, **params):
        """
        Catboost Classifier wrapper
        """
        self.model = catboost.CatBoostClassifier()
        self.classes_ = None
        self.trained = False
        self.callbacks = None
        self.verbose = 0
        self.early_stopping_rounds = 100
        self.use_best_model = True
        self.set_params(**params)

    def set_params(self, **params):
        # Add defaults if not present
        for k, v in self.default_params.items():
            if k not in params.keys():
                params[k] = v

        # Take out fit settings
        if "early_stopping_rounds" in params:
            self.early_stopping_rounds = params.pop("early_stopping_rounds")
        if "use_best_model" in params:
            self.use_best_model = params.pop("use_best_model")
        if "verbose" in params:
            self.verbose = params.pop("verbose")

        # Update params
        self.params = params
        self.model.set_params(**params)
        return self

    def get_params(self, *args, **kwargs):
        return self.model.get_params(*args, **kwargs)

    def fit(self, X, y, *args, **kwargs):
        # Split data
        train_x, test_x, train_y, test_y = train_test_split(
            X, y, stratify=y, test_size=0.1
        )

        # Set Attributes
        self.classes_ = np.unique(y)

        # Train model
        self.model.fit(
            train_x,
            train_y,
            eval_set=[(test_x, test_y)],
            verbose=self.verbose,
            early_stopping_rounds=self.early_stopping_rounds,
            use_best_model=self.use_best_model,
        )

        # Set trained
        self.trained = True

    def predict(self, X, *args, **kwargs):
        return self.model.predict(X, *args, **kwargs).reshape(-1)

    def predict_proba(self, X, *args, **kwargs):
        return self.model.predict_proba(X, *args, **kwargs)



#  Copyright (c) 2022 by Amplo.

import json
import os
import warnings
from datetime import datetime
from pathlib import Path

from azure.storage.blob import BlobServiceClient

__all__ = ["AzureSynchronizer"]


class AzureSynchronizer:
    def __init__(
        self,
        connection_string_name="AZURE_STORAGE_STRING",
        container_client_name="amploplatform",
        verbose=0,
    ):
        """
        Connector to Azure storage blob for downloading data that is stored
        in Amplo`s data storage fashion.

        Parameters
        ----------
        connection_string_name : str
        container_client_name : str
        verbose : int
        """
        client = BlobServiceClient.from_connection_string(
            os.getenv(connection_string_name)
        )
        self.container = client.get_container_client(container_client_name)
        # TODO: Use Amplo`s logging instead of a print function
        self.verbose = int(verbose)
        self._metadata_filename = ".metadata"
        self._str_time_format = "%Y-%m-%d %H:%M:%S:%z"

    def get_dir_paths(self, path=None):
        """
        Get all directories that are direct children of given directory (``path``).

        Parameters
        ----------
        path : str or Path, optional
            Path to search for directories.
            If not provided, searches in root `/`.

        Returns
        -------
        list of str
        """
        if path is not None:
            # Provide a slash from right
            path = f"{Path(path)}/"
        dirs = [
            b.name for b in self.container.walk_blobs(path) if str(b.name).endswith("/")
        ]
        return dirs

    def get_filenames(self, path, with_prefix=False, sub_folders=False):
        """
        Get all files that are direct children of given directory (``path``).

        Parameters
        ----------
        path : str or Path
            Path to search for files
        with_prefix : bool
            Whether to fix the prefix of the files
        sub_folders : bool
            Whether to search also for files inside sub-folders

        Returns
        -------
        list of str
        """
        # Provide a slash from right
        path = f"{Path(path)}/"

        # List files
        if sub_folders:
            files = [
                f.name
                for f in self.container.walk_blobs(path, delimiter="")
                if "." in f.name
            ]
        else:
            files = [
                f.name
                for f in self.container.walk_blobs(path, delimiter="")
                if f.name.count("/") == path.count("/") and "." in f.name
            ]

        # Fix prefix
        if not with_prefix:
            files = [f[len(path) :] for f in files]

        # Remove empties
        if "" in files:
            files.remove("")

        return files

    def sync_files(self, blob_dir, local_dir):
        """
        Download all files inside blob directory and store it to the local directory.

        Additionally, creates a file `.metadata` that stores additional info about
        synchronization such as blob directory path and last modification date.

        Parameters
        ----------
        blob_dir : str or Path
            Search directory (download)
        local_dir : str or Path
            Local directory (store)

        Returns
        -------
        found_new_data : bool
            Whether new data has been downloaded

        Notes
        -----
        The data in the `.metadata` file is currently only used for checking the last
        modification date, thus telling whether files have to be downloaded / updated.
        """
        # Set up paths
        blob_dir = Path(blob_dir)
        local_dir = Path(local_dir)
        metadata_dir = local_dir

        # Warn when names don't match
        if blob_dir.name != local_dir.name:
            warnings.warn(
                f"Name mismatch detected. {blob_dir.name} != {local_dir.name}"
            )

        # Skip "Random"
        if blob_dir.name == "Random":
            warnings.warn(f"Skipped synchronization from {blob_dir}")
            return False

        # Set up metadata
        blob_dir_properties = self.container.get_blob_client(
            str(blob_dir)
        ).get_blob_properties()
        metadata = dict(
            blob_dir=str(blob_dir),
            container=blob_dir_properties.container,
            last_modified=blob_dir_properties.last_modified,
            new_files=[],
        )

        # Load local metadata from previous synchronization
        local_metadata = self.load_local_metadata(metadata_dir, not_exist_ok=True)
        last_updated = local_metadata.get(
            "last_modified", datetime(year=1900, month=1, day=1).astimezone()
        )

        # Read & write all files
        for file in self.get_filenames(blob_dir):

            # Create directory only if files are found
            local_dir.mkdir(parents=True, exist_ok=True)

            # Get file blob
            blob = self.container.get_blob_client(str(blob_dir / file))
            blob_properties = blob.get_blob_properties()

            # Download and save if file is new or modified
            file_created: datetime = blob_properties.creation_time
            file_last_modified: datetime = blob_properties.last_modified
            if file_last_modified > last_updated:
                # Write file
                with open(str(local_dir / file), "wb") as f:
                    f.write(blob.download_blob().readall())
                # Match timestamps of local file with blob
                os.utime(
                    str(local_dir / file),
                    (file_created.timestamp(), file_last_modified.timestamp()),
                )
                # Increment
                metadata["new_files"] += [str(file)]

        # Check whether found new data
        found_new_data = metadata["last_modified"] > last_updated
        if found_new_data:
            # Store metadata
            self._dump_local_metadata(metadata, metadata_dir)

        return found_new_data

    # --- Utilities ---

    def load_local_metadata(self, local_dir, *, not_exist_ok=False):
        metadata_path = Path(local_dir) / self._metadata_filename
        if metadata_path.exists():
            # Get and check local metadata
            metadata = json.load(open(str(metadata_path), "r"))
            assert isinstance(metadata, dict), f"Damaged metadata in {metadata_path}"
            # Convert string to datetime object
            metadata["last_modified"] = datetime.strptime(
                metadata["last_modified"], self._str_time_format
            )
            return metadata
        elif not_exist_ok:
            return dict()
        else:
            raise FileNotFoundError(f"File {metadata_path} does not exist")

    def _dump_local_metadata(self, metadata, local_dir):
        metadata_path = Path(local_dir) / self._metadata_filename
        # Check metadata keys
        assert {"blob_dir", "container", "new_files", "last_modified"}.issubset(
            metadata.keys()
        ), "Invalid metadata"
        # Make datetime object JSON serializable
        metadata["last_modified"] = datetime.strftime(
            metadata["last_modified"], self._str_time_format
        )
        # Dump
        json.dump(metadata, open(str(metadata_path), "w"))



#  Copyright (c) 2022 by Amplo.

import logging
from datetime import datetime

import colorlog

__all__ = ["logger"]


_nameToLevel = {
    "CRITICAL": logging.CRITICAL,
    "FATAL": logging.FATAL,
    "ERROR": logging.ERROR,
    "WARN": logging.WARNING,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
    "NOTSET": logging.NOTSET,
}


def _check_logging_level(level):
    """
    Check input whether it is a valid logging level

    Parameters
    ----------
    level : int or str
        Logging level

    Returns
    -------
    int
        (valid) logging level

    Raises
    ------
    ValueError
        If the given logging level is not valid
    """
    # Inspired by logging/__init__.py
    if isinstance(level, int):
        if level not in set(_nameToLevel.values()):
            raise ValueError("Unknown level: %r" % level)
        rv = level
    elif str(level) == level:
        if level not in _nameToLevel:
            raise ValueError("Unknown level: %r" % level)
        rv = _nameToLevel[level]
    else:
        raise TypeError("Level not an integer or a valid string: %r" % level)
    return rv


class TimeFilter(logging.Filter):
    def filter(self, record):
        # Check if previous logged
        if not hasattr(self, "last"):
            self.last = record.relativeCreated

        # Calc & add delta
        delta = datetime.fromtimestamp(
            record.relativeCreated / 1000.0
        ) - datetime.fromtimestamp(self.last / 1000.0)
        record.relative = f"{(delta.seconds + delta.microseconds / 1e6):.2f}"

        # Update last
        self.last = record.relativeCreated

        return True


# Get custom logger
logger = logging.getLogger("AmploML")
logger.setLevel("INFO")

# Set console handler
console_formatter = colorlog.ColoredFormatter(
    "%(blue)s[%(name)s]%(log_color)s[%(levelname)s] %(white)s%(message)s "
    "%(light_black)s<%(filename)s:%(lineno)d> (%(relative)ss)"
)
console_handler = logging.StreamHandler()
console_handler.setLevel("INFO")
console_handler.setFormatter(console_formatter)
logger.addHandler(console_handler)

# Set file handler
file_formatter = logging.Formatter(
    "[%(name)s][%(levelname)s] %(message)s <%(filename)s:%(lineno)d> (%(relative)ss)"
)
file_handler = logging.FileHandler("AutoML.log", mode="a")
file_handler.setLevel("INFO")
file_handler.setFormatter(file_formatter)
logger.addHandler(file_handler)

# Add TimeFilter
[handler.addFilter(TimeFilter()) for handler in logger.handlers]

# Capture warnings from `warnings.warn(...)`
logging.captureWarnings(True)
py_warnings_logger = logging.getLogger("py.warnings")
warnings_formatter = colorlog.ColoredFormatter(
    "%(white)s[%(name)s] %(log_color)s%(levelname)s: %(message)s"
)
warnings_handler = logging.StreamHandler()
warnings_handler.setLevel("WARNING")
warnings_handler.setFormatter(warnings_formatter)
py_warnings_logger.addHandler(warnings_handler)


def remove_file_handler():
    logger.removeHandler(file_handler)



#  Copyright (c) 2022 by Amplo.

import catboost
from sklearn.model_selection import train_test_split


class CatBoostRegressor:
    _estimator_type = "regressor"
    default_params = {
        "verbose": 0,
        "n_estimators": 1000,
        "allow_writing_files": False,
    }

    def __init__(self, **params):
        """
        Catboost Regressor wrapper
        """
        self.model = catboost.CatBoostRegressor()
        self.trained = False
        self.callbacks = None
        self.verbose = 0
        self.early_stopping_rounds = 100
        self.set_params(**params)

    def set_params(self, **params):
        # Add default if necessary
        for k, v in self.default_params.items():
            if k not in params:
                params[k] = v

        # Remove fit options
        if "early_stopping_rounds" in params:
            self.early_stopping_rounds = params.pop("early_stopping_rounds")
        if "verbose" in params:
            self.verbose = params.pop("verbose")

        # Update model & class
        self.model.set_params(**params)
        self.params = params
        return self

    def get_params(self, **args):
        return self.model.get_params(**args)

    def fit(self, x, y):
        # Split data
        train_x, test_x, train_y, test_y = train_test_split(x, y, test_size=0.1)

        # Train model
        self.model.fit(
            train_x,
            train_y,
            eval_set=[(test_x, test_y)],
            verbose=self.verbose,
            early_stopping_rounds=self.early_stopping_rounds,
        )

        # Set trained
        self.trained = True

    def predict(self, x, *args, **kwargs):
        return self.model.predict(x, *args, **kwargs).reshape(-1)



#  Copyright (c) 2022 by Amplo.

import json
import os
from pathlib import Path
from datetime import datetime, timedelta

import joblib
import numpy as np
import pandas as pd
import pytest
from sklearn.metrics import log_loss, r2_score

from amplo import Pipeline
from amplo.utils.testing import make_interval_data
from tests import make_data, rmtree


@pytest.fixture(scope="class", params=["regression", "classification"])
def make_mode(request, target="target"):
    mode = request.param
    request.cls.mode = mode
    request.cls.target = target
    request.cls.data = make_data(mode=mode, target=target)
    yield


@pytest.mark.usefixtures("make_mode")
class TestPipelineByMode:
    def test_mode_detector(self):
        # Numerical test
        pipeline = Pipeline(no_dirs=True, target=self.target)
        pipeline._read_data(self.data)._mode_detector()
        assert pipeline.mode == self.mode

        # Categorical test
        if self.mode == "classification":
            # Convert to categorical
            data = self.data.copy()
            data[self.target] = [f"Class_{v}" for v in data[self.target].values]
            # Detect mode
            pipeline = Pipeline(no_dirs=True, target=self.target)
            pipeline._read_data(data)._mode_detector()
            assert pipeline.mode == self.mode

    def test_stacking(self):
        # Set stacking model
        if self.mode == "classification":
            stacking_model = "StackingClassifier"
        else:
            stacking_model = "StackingRegressor"
        # Fit pipeline
        pipeline = Pipeline(
            target=self.target,
            grid_search_type=None,
            stacking=True,
            extract_features=False,
        )
        pipeline.fit(self.data, model=stacking_model)

    def test_dir_and_settings(self):
        pipeline = Pipeline(
            target=self.target,
            mode=self.mode,
            objective="r2" if self.mode == "regression" else "neg_log_loss",
            n_grid_searches=0,
            plot_eda=False,
            process_data=False,
            extract_features=False,
            document_results=False,
        )
        pipeline.fit(self.data)

        # Test Directories
        assert os.path.exists("Auto_ML")
        assert os.path.exists("Auto_ML/Data")
        assert os.path.exists("Auto_ML/Features")
        assert os.path.exists("Auto_ML/Production")
        assert os.path.exists("Auto_ML/Documentation")
        assert os.path.exists("Auto_ML/Results.csv")

        if self.mode == "classification":
            # Pipeline Prediction
            prediction = pipeline.predict_proba(self.data)
            assert log_loss(self.data[self.target], prediction) > -1

        elif self.mode == "regression":
            # Test data handling
            c, _ = pipeline.convert_data(self.data.drop(self.target, axis=1))
            x = pipeline.x[pipeline.settings["features"]]
            y = self.data[self.target]
            assert np.allclose(
                c.values, x.values
            ), f"Inconsistent X: max diff: {np.max(abs(c.values - x.values)):.2e}"
            assert np.allclose(y, pipeline.y), "Inconsistent y-data"

            # Pipeline Prediction
            prediction = pipeline.predict(self.data)
            assert len(prediction.shape) == 1
            assert r2_score(self.data[self.target], prediction) > 0.75

        else:
            raise ValueError(f"Invalid mode {self.mode}")

        # Settings prediction
        settings = json.load(open("Auto_ML/Production/v1/Settings.json", "r"))
        model = joblib.load("Auto_ML/Production/v1/Model.joblib")
        p = Pipeline(no_dirs=True)
        p.load_settings(settings)
        p.load_model(model)
        if self.mode == "classification":
            assert np.allclose(p.predict_proba(self.data), prediction)
        else:
            assert np.allclose(p.predict(self.data), prediction)

        # Check settings
        assert "pipeline" in settings
        assert "version" in settings
        assert "model" in settings
        assert "amplo_version" in settings
        assert "params" in settings
        assert "drift_detector" in settings
        assert "features" in settings
        assert "validation" in settings
        assert "data_processing" in settings


class TestPipeline:
    def test_drift(self):
        # todo connect with logger
        pytest.skip("This test is not yet implemented")

    def test_all_models(self):
        # TODO
        pytest.skip("This test is not yet implemented")

    def test_interval_analyzer(self):
        """
        Use interval analyzer when a folder with logs is presented to pipeline.
        """
        # Create dummy data
        data_dir = "./test_data"
        make_interval_data(directory=data_dir)

        # Pipeline
        pipeline = Pipeline(n_grid_searches=0)
        pipeline.fit(data_dir)

        # Check if IntervalAnalyser is fitted
        assert pipeline.interval_analyser.is_fitted, "IntervalAnalyser was not fitted"

        # Check data handling
        assert Path(
            "Auto_ML/Data/Interval_Analyzed_v1.csv"
        ).exists(), "Interval-analyzed data was not properly stored"
        assert list(pipeline.data.index.names) == ["log", "index"], "Index is incorrect"
        # TODO: add checks for settings

        # Remove dummy data
        rmtree(data_dir, must_exist=True)

    def test_with_time(self):
        x = np.vstack(
            (
                [datetime(2020, 1, 1) + timedelta(days=i) for i in range(12)],
                np.linspace(0, 10, 12),
            )
        ).T
        y = np.linspace(0, 10, 12)
        pipeline = Pipeline(n_grid_searches=0)
        pipeline.fit(x, y)

        # Do the same, but with temporal
        multi_index = pd.MultiIndex.from_product(
            [[1, 2, 3], [1, 2, 3, 4]], names=["log", "index"]
        )
        df = pd.DataFrame(x, index=multi_index)
        pipeline = Pipeline(n_grid_searches=0)
        pipeline.fit(df, y)



#  Copyright (c) 2022 by Amplo.

"""
Observer for checking data.

This part of code is strongly inspired by [1].

References
----------
[1] E. Breck, C. Shanging, E. Nielsen, M. Salib, D. Sculley (2017).
The ML test score: A rubric for ML production readiness and technical debt
reduction. 1123-1132. 10.1109/BigData.2017.8258038.
"""

import json

import numpy as np
import pandas as pd

from amplo.observation._base import PipelineObserver, _report_obs
from amplo.utils.logging import logger
from amplo.utils.metrics import levenshtein_distance

__all__ = ["DataObserver"]


class DataObserver(PipelineObserver):
    """
    Data observer before pushing to production.

    Machine learning systems differ from traditional software-based systems in
    that the behavior of ML systems is not specified directly in code but is
    learned from data. Therefore, while traditional software can rely on unit
    tests and integration tests of the code, here we attempt to add a sufficient
    set of tests of the data.

    The following tests are included:
        1. Feature columns should not be monotonically in-/decreasing.
        2. Feature columns should not be sensitive in minority classes
    """

    _obs_type = "data_observation"

    def observe(self):
        self.check_monotonic_columns()
        self.check_minority_sensitivity()
        self.check_extreme_values()
        self.check_categorical_mismatch()

    @_report_obs
    def check_monotonic_columns(self):
        """
        Checks whether any column is monotonically in- or decreasing.

        When data is multi-indexed it checks monotonicity per group (level 0).

        Returns
        -------
        status_ok : bool
            Observation status. Indicates whether a warning should be raised.
        message : str
            A brief description of the observation and its results.
        """
        logger.info("Checking data for monotonic columns (often indices).")
        x_data = pd.DataFrame(self.x)
        numeric_data = x_data.select_dtypes(include=np.number)

        # Make multi-indexed if not already
        if len(numeric_data.index.names) == 1:
            numeric_data.index = pd.MultiIndex.from_product([[0], numeric_data.index])

        # Sort index from the innermost to outermost index level since it is shuffled
        # in classification.
        for axis in list(range(len(numeric_data.index.names)))[::-1]:
            numeric_data.sort_index(axis=axis, inplace=True)

        monotonic_columns = []
        for col in numeric_data.columns:
            grouped_series = (
                numeric_data[col]
                .groupby(level=0)
                .apply(lambda group: group.interpolate(limit_directions="both"))
                .groupby(level=0)
            )
            is_monotonic = (
                grouped_series.apply(lambda group: group.is_monotonic)
                | grouped_series.apply(lambda group: group.is_monotonic_decreasing)
            ).any()
            is_constant = grouped_series.apply(lambda group: group.nunique() == 1).any()
            if is_monotonic and not is_constant:
                monotonic_columns.append(col)

        status_ok = not monotonic_columns
        message = (
            f"{len(monotonic_columns)} columns are monotonically in- or "
            f"decreasing. More specifically: {monotonic_columns}"
        )
        return status_ok, message

    @_report_obs
    def check_minority_sensitivity(self):
        """
        Checks whether the data has sensitivities towards minority classes.

        Minority sensitivity is a concept where a signal is present in a small subsample
        of a minority class. As this minority class is potentially not well-covered,
        the small subsample should not be indicative to identify the class or vice versa

        This is analysed by a simple discrete distribution of 10 bins. Minority
        sensitivity is defined as:
        - Bin contains < 2% of total data
        - Bin contains to one class
        - Bin contains > 20% of that class
        (-> class needs to be 10% of data or smaller)

        Note: Only supports numeric columns

        Returns
        -------
        status_ok : bool
            Observation status. Indicates whether a warning should be raised.
        message : str
            A brief description of the observation and its results.
        """
        logger.info("Checking data for minority sensitive columns.")
        minority_sensitive = []

        for key in self.x.keys():
            if not pd.api.types.is_numeric_dtype(self.x[key]):
                # Todo implement for categorical columns
                continue

            # Make bins
            counts, edges = np.histogram(self.x[key].fillna(0), bins=10)

            # Check if a minority group is present
            minority_size = min([c for c in counts if c != 0])
            if minority_size > len(self.x) // 50:
                continue

            # If present, check the labels
            bin_indices = np.digitize(self.x[key], bins=edges)
            for bin_ind in np.where(counts == minority_size)[0]:
                minority_indices = np.where(bin_indices == bin_ind + 1)[0]

                # No minority if spread across labels
                if self.y.iloc[minority_indices].nunique() != 1:
                    continue

                # Not sensitive if only a fraction of the label
                if (
                    len(minority_indices)
                    > (self.y == self.y.iloc[minority_indices[0]]).sum() // 5
                ):
                    minority_sensitive.append(key)

        status_ok = not minority_sensitive
        message = (
            f"{len(minority_sensitive)} columns have minority sensitivity. "
            "Consider to remove them or add data."
            f" More specifically: {minority_sensitive}."
        )
        return status_ok, message

    @_report_obs
    def check_extreme_values(self):
        """
        Checks whether extreme values are present.

        Returns
        -------
        status_ok : bool
            Observation status. Indicates whether a warning should be raised.
        message : str
            A brief description of the observation and its results.
        """
        logger.info("Checking data for extreme values.")
        extreme_values = []

        for key in self.x.keys():
            if not pd.api.types.is_numeric_dtype(self.x[key]):
                # Todo implement for categorical columns
                continue

            if self.x[key].abs().max() > 1000:
                extreme_values.append(key)

        status_ok = not extreme_values
        message = (
            f"{len(extreme_values)} columns have values > 1000. "
            f" More specifically: {extreme_values}."
        )
        return status_ok, message

    @_report_obs
    def check_categorical_mismatch(self):
        """
        Checks whether categorical variables are mismatched

        For example "New York" and "new york". We do this with a simple regex, removing
        all special characters and lowercasing the category.

        Returns
        -------
        status_ok : bool
            Observation status. Indicates whether a warning should be raised.
        message : str
            A brief description of the observation and its results.
        """
        logger.info("Checking data for categorical mismatches.")
        categorical_mismatches = []

        # Get dummy information from pipeline
        for feature, variants in self._pipe.data_processor.dummies.items():

            # Check if one is similar
            for i, variant_x in enumerate(variants):
                for j, variant_y in enumerate(variants):
                    # We only need to compare all once
                    if j >= i:
                        continue

                    # Remove feature from variants
                    variant_x = variant_x[len(feature) + 1 :]
                    variant_y = variant_y[len(feature) + 1 :]

                    # Get distance (normalized)
                    distance = levenshtein_distance(variant_x, variant_y) / max(
                        len(variant_x), len(variant_y)
                    )

                    # Compare
                    if 0 < distance <= 0.2:
                        categorical_mismatches.append({feature: [variant_x, variant_y]})

        status_ok = not categorical_mismatches
        message = (
            f"{len(categorical_mismatches)} categorical columns have mismatching."
            f"categories. More specifically: {json.dumps(categorical_mismatches)}."
        )
        return status_ok, message



#  Copyright (c) 2022 by Amplo.

from amplo.api.storage import AzureSynchronizer
from tests.unit.api import TestAPI


class TestStorage(TestAPI):

    def test_get_dir_paths(self):
        blob_dir = 'Demo/Charger 75kW/data'
        sync = AzureSynchronizer()
        dirs_actual = sync.get_dir_paths(blob_dir)
        dirs_target = [f'{blob_dir}/{folder}/' for folder in ('Diagnostics', 'Healthy', 'Unlabelled')]
        assert set(dirs_actual) == set(dirs_target), \
            'The function\'s output did not match the expected output'

    def test_get_filenames(self):
        blob_dir = 'Demo/Charger 75kW/data/Diagnostics/xx'
        sync = AzureSynchronizer()
        files_actual = sync.get_filenames(blob_dir)
        files_target = ['beth10.csv', 'long-data.csv', 'long-dataxx.csv', 'long-dataxxy.csv']
        assert set(files_actual) == set(files_target), \
            'The function\'s output did not match the expected output'

    def test_sync_files(self):
        blob_dir = 'Demo/Charger 75kW/data/Diagnostics/xx'
        local_dir = self.sync_dir / 'Demo/Charger 75kW/Diagnostics/data/xx'
        # Set up API and synch data
        sync = AzureSynchronizer()
        sync.sync_files(blob_dir, local_dir)
        # Assert that files have been synchronized
        blob_files = sync.get_filenames(blob_dir)
        local_files = [path.name for path in local_dir.glob('*')]
        assert set(blob_files).issubset(local_files), 'Synchronization has failed. Content is not equal...'



from functools import lru_cache

__all__ = ["levenshtein_distance"]


def levenshtein_distance(a, b):
    """
    Calculates the levenshtein distance between two words.

    parameters
    ----------
    a : str
        First word
    b : str
        Second word

    returns
    -------
    distance : int
    """

    @lru_cache(None)  # for memorization
    def min_dist(s1, s2):

        # Deal with different lengths
        if s1 == len(a) or s2 == len(b):
            return len(a) - s1 + len(b) - s2

        # Continue when they're the same
        if a[s1] == b[s2]:
            return min_dist(s1 + 1, s2 + 1)

        # If not, add point and continue
        dist = min(
            min_dist(s1, s2 + 1),  # Insert character
            min_dist(s1 + 1, s2),  # Delete character
            min_dist(s1 + 1, s2 + 1),  # Replace character
        )

        # Damerau
        if (
            s1 < len(a) - 1
            and s2 < len(b) - 1
            and a[s1] == b[s2 + 1]
            and a[s1 + 1] == b[s2]
        ):
            dist = min(dist, min_dist(s1 + 2, s2 + 2))
        return dist + 1

    return min_dist(0, 0)



#  Copyright (c) 2022 by Amplo.

"""
Base class used to build new observers.
"""

import abc
import warnings
from typing import TYPE_CHECKING, Dict, List, Union

import numpy as np
from sklearn.metrics import get_scorer

if TYPE_CHECKING:
    from amplo import Pipeline

__all__ = ["BaseObserver", "PipelineObserver", "ProductionWarning", "_report_obs"]


class ProductionWarning(RuntimeWarning):
    """
    Warning for suspicions before moving to production.
    """


class BaseObserver(abc.ABC):
    """
    Abstract base class to build new observers.

    Subclass this class.

    Attributes
    ----------
    observations : list of dict
        A list of observations.  Each observation is a dictionary containing the
        keys `type` (str), `name` (str), `status_ok` (bool) and `description`
        (str) - with corresponding dtypes.
    """

    def __init__(self):
        self.observations: List[Dict[str, Union[str, bool]]] = []

    def report_observation(self, typ, name, status_ok, message):
        """
        Report an observation to the observer.

        An observation will trigger a warning when `status_ok` is false.

        Parameters
        ----------
        typ : str
            Observation type.
        name : str
            Observation name.
        status_ok : bool
            Observation status. If false, a warning will be triggered.
        message : str
            A brief description of the observation and its results.
        """
        # Check input
        if not isinstance(typ, str):
            raise ValueError("Invalid dtype for observation type.")
        if not isinstance(name, str):
            raise ValueError("Invalid dtype for observation name.")
        if not isinstance(status_ok, (bool, np.bool_)):
            raise ValueError("Invalid dtype for observation status.")
        if not isinstance(message, str):
            raise ValueError("Invalid dtype for observation message.")

        # Trigger warning when status is not okay
        if not status_ok:
            msg = (
                "A production observation needs inspection. Please evaluate "
                f"why a warning was triggered from `{typ}/{name}`. "
                f"Warning message: {message}"
            )
            warnings.warn(ProductionWarning(msg))

        # Add observation to list
        obs = {"typ": typ, "name": name, "status_ok": status_ok, "message": message}
        self.observations.append(obs)

    @abc.abstractmethod
    def observe(self):
        """
        Observe the data, model, ...

        Observations should be reported via `self.report_observation()`.
        """


class PipelineObserver(BaseObserver, metaclass=abc.ABCMeta):
    """
    Extension of ``BaseObserver``.

    Unifies behavior of class initialization.

    Parameters
    ----------
    pipeline : Pipeline
        The amplo pipeline object that will be observed.

    Class Attributes
    ----------------
    _obs_type : str
        Name of the observation.
    CLASSIFICATION : str
        Name for a classification mode.
    REGRESSION : str
        Name for a regression mode.
    """

    _obs_type = None
    CLASSIFICATION = "classification"
    REGRESSION = "regression"

    def __init__(self, pipeline: "Pipeline"):
        super().__init__()

        if not type(pipeline).__name__ == "Pipeline":
            raise ValueError("Must be an Amplo pipeline.")

        self._pipe = pipeline

    @property
    def obs_type(self) -> str:
        """
        Name of the observation type.
        """
        if not self._obs_type or not isinstance(self._obs_type, str):
            raise AttributeError("Class attribute `_obs_type` is not set.")
        return self._obs_type

    @property
    def mode(self):
        return self._pipe.mode

    @property
    def scorer(self):
        return get_scorer(self._pipe.objective)

    @property
    def x(self):
        return self._pipe.x

    @property
    def y(self):
        return self._pipe.y


def _report_obs(func):
    """
    Decorator for checker function in observer class.

    Parameters
    ----------
    func : function
        The class method that shall report an observation. It must return the
        observation status (bool) and its message (str).

    Returns
    -------
    decorator
    """

    def report(self: PipelineObserver, *args, **kwargs):
        assert isinstance(self, PipelineObserver)
        status_ok, message = func(self, *args, **kwargs)
        self.report_observation(self.obs_type, func.__name__, status_ok, message)

    return report



#  Copyright (c) 2022 by Amplo.

"""
Observer for checking production readiness.
"""
from time import time

from amplo.observation._base import PipelineObserver, _report_obs

__all__ = ["ProductionObserver"]


class ProductionObserver(PipelineObserver):
    """
    Observer before putting to production.

    Testing and monitoring are key considerations for ensuring the production-
    readiness of an ML system, and for reducing technical debt of ML systems.

    Parameters
    ----------
    pipeline : Pipeline
        The amplo pipeline object that will be observed.
    """

    def __init__(self, pipeline):
        super().__init__(pipeline)
        # TODO read below
        # We need this to be executed every time a prediction is made.
        # Should probably use these as wrappers for the predict & predict_proba
        # function to collect the statistics. For now, this class is not used.

    def observe(self):
        # self.check_prediction_latency()
        pass

    @_report_obs
    def check_prediction_latency(self, threshold=0.1):
        """
        Check the latency of predicting a single sample.

        If it takes longer than 100ms, something is wrong.

        Parameters
        ----------
        threshold : float
            Threshold for latency (in s).

        Returns
        -------
        status_ok : bool
            Observation status. Indicates whether a warning should be raised.
        message : str
            A brief description of the observation and its results.
        """
        prediction_latencies = []
        for repeat in range(10):
            t_start = time()
            self.fitted_model.predict(self.x.sample(n=1))
            prediction_latencies.append(time() - t_start)

        status_ok = not any(latency > threshold for latency in prediction_latencies)
        message = (
            f"The latency of predicting a single sample took longer than "
            f"{threshold * 1e3:.0f}ms. Latencies: {prediction_latencies}."
        )

        return status_ok, message



#  Copyright (c) 2022 by Amplo.

import numpy as np


class BaseClassifier:
    _estimator_type = "classifier"

    def __init__(self, default_params=None, **params):
        assert isinstance(
            default_params, dict
        ), "Provided default parameters not of type dict"
        assert isinstance(params, dict), "Provided parameters not of type dict"
        self.default_params = default_params
        self.model = None
        self.has_predict_proba = False
        self.trained = False
        self.classes_ = None
        self.callbacks = None
        self.params = params if params is not None else self.default_params
        for key in [k for k in self.default_params if k not in self.params]:
            self.params[key] = self.default_params[key]

    def get_params(self):
        return self.model.get_params()

    def set_params(self, **params):
        self.model.set_params(**params)
        return self

    def reset_weights(self):
        pass

    def predict(self, X):
        return self.model.predict(X)

    def predict_proba(self, X):
        if self.has_predict_proba:
            return self.model.predict_proba(X)
        else:
            raise AttributeError(
                "{} has no predict_proba".format(type(self.model).__name__)
            )

    def score(self, X, y):
        return self.model.score(X, y)

    def fit(self, X, y):
        self.classes_ = np.unique(y)
        self.model.fit(X, y)
        self.trained = True
        return self



#  Copyright (c) 2022 by Amplo.

"""
Implements the basic behavior of feature processing.
"""

from copy import deepcopy
from typing import List
from warnings import warn

import numpy as np
import pandas as pd
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

from amplo.base import BaseTransformer, LoggingMixin
from amplo.utils import check_dtypes
from amplo.utils.data import clean_keys

__all__ = [
    "sanitize_series",
    "sanitize_dataframe",
    "BaseFeatureProcessor",
    "BaseFeatureExtractor",
]


def sanitize_series(series):
    """
    Sanitizes series data.

    - Replaces `np.inf` and `np.nan` with zeros.
    - Clips min and max values.
    - Converts all floats and ints to `np.float32` and `np.int32`, respectively.

    Parameters
    ----------
    series : pd.Series
        Data to be sanitized.

    Returns
    -------
    pd.Series
        Sanitized data.
    """
    dtype = series.dtype
    if pd.api.types.is_datetime64_any_dtype(series) or not np.issubdtype(
        dtype, np.number
    ):
        return series

    # Get info for 32-bit data
    if np.issubdtype(dtype, np.floating):
        info = np.finfo(np.float32)
    elif np.issubdtype(dtype, np.integer):
        info = np.iinfo(np.int32)
    else:
        raise TypeError(f"Invalid dtype for series data: {series.dtype}")
    # Sanitize
    lower = max(info.min, -1e12)
    upper = min(info.max, 1e12)
    series = series.fillna(0).clip(lower, upper).astype(info.dtype)
    return series


def sanitize_dataframe(frame):
    """
    Sanitizes each column.

    Parameters
    ----------
    frame : pd.DataFrame
        Data to be sanitized.

    Returns
    -------
    pd.DataFrame
        Sanitized data.
    """
    return frame.apply(sanitize_series, axis=0)


class BaseFeatureProcessor(BaseTransformer, LoggingMixin):
    """
    Base class for feature processors.

    Parameters
    ----------
    mode : str, optional, default: "notset"
        Model mode: {"classification", "regression", "notset"}.
    verbose : int
        Verbosity for logger.
    """

    def __init__(self, mode="notset", verbose=0):
        BaseTransformer.__init__(self)
        LoggingMixin.__init__(self, verbose=verbose)

        check_dtypes([("mode", mode, (str, type(None)))])

        if isinstance(mode, str):
            mode = mode.lower()
        if mode == "notset" or mode is None:
            # Allows the class to be initialized without specifying the mode
            mode = "notset"
        elif mode == "multiclass":
            mode = "classification"
        elif mode not in ("classification", "regression"):
            raise ValueError("Invalid `mode` argument.")

        self.mode = mode

    @staticmethod
    def _check_x(x, copy=True, sanitize=True):
        """
        Check and sanitize x data.

        Parameters
        ----------
        x : pd.DataFrame
        copy : bool
        sanitize : bool

        Returns
        -------
        pd.DataFrame
        """
        check_dtypes([("x", x, pd.DataFrame)])

        x = deepcopy(x) if copy else x  # make copy

        # Stringify names
        x = clean_keys(x)
        assert not [  # important for `raw_features_` attribute to work properly
            col for col in x.columns if "__" in col
        ], "Cleaned column names should not have double underscores in it."
        if any(x.columns.duplicated()):
            raise ValueError("Feature column names are not unique.")

        # Sanitize values
        x_checked = sanitize_dataframe(x) if sanitize else x
        return x_checked

    @staticmethod
    def _check_y(y, copy=True, sanitize=True):
        """
        Check and sanitize x data.

        Parameters
        ----------
        y : pd.Series
        copy : bool
        sanitize : bool

        Returns
        -------
        pd.Series
        """
        check_dtypes([("y", y, pd.Series)])

        y = deepcopy(y) if copy else y  # make copy
        y.name = str(y.name) if y.name else "target"  # stringify name

        # Sanitize values
        y_checked = sanitize_series(y) if sanitize else y
        return y_checked

    @staticmethod
    def _check_x_y(x, y, copy=True, sanitize=True):

        # Pre-check
        x = BaseFeatureProcessor._check_x(x=x, copy=copy, sanitize=sanitize)
        y = BaseFeatureProcessor._check_y(y=y, copy=copy, sanitize=sanitize)

        # Check integrity
        if len(x) != len(y):
            raise ValueError("Length of `x` and `y` data does not match.")
        if not all(x.index == y.index):
            warn(
                "Indices of x and y data do not match. Setting x index as default.",
                UserWarning,
            )
            y.index = x.index

        return x, y

    def _fit_transform(self, x, y=None, **fit_params):
        """
        Fit and transform data.

        It's faster calling this function instead of `_fit` and `_transform` separately.

        Parameters
        ----------
        x : numpy.ndarray or pandas.DataFrame
            Checked feature data to fit.
        y : numpy.ndarray or pandas.Series
            Checked target data to fit.

        Returns
        -------
        pandas.DataFrame
            Transformed version of x.
        """
        raise NotImplementedError("Abstract method.")

    def _transform(self, x, y=None):
        raise NotImplementedError("Abstract method.")

    def fit(self, x, y=None, **fit_params):
        # Feature processors always have to transform the data in order to fit. So it
        # just makes sense to wrap the main behavior in the fit_transform function and
        # call that in here.
        self.fit_transform(x=x, y=y, **fit_params)
        return self

    def fit_transform(self, x, y=None, **fit_params) -> pd.DataFrame:
        # This is an optimized version for BaseFeatureProcessor's fit_transform.

        # If fit is called, estimator is reset, including fitted state
        self.reset()

        # Pass to inner fit
        x_out = self._fit_transform(x, y, **fit_params)
        if not isinstance(x_out, pd.DataFrame):
            msg = (
                "The '_fit' function of classes that inherit from BaseFeatureProcessor "
                "must output a DataFrame. If other behavior is desired, please adjust "
                "the `fit_transform` accordingly."
            )
            raise ValueError(msg)

        # This should happen last: fitted state is set to True
        self._is_fitted = True

        return x_out


class BaseFeatureExtractor(BaseFeatureProcessor):
    """
    Base class for feature extractors.

    Fitted attributes:
        Extracted feature names are stored in "features_".

    Parameters
    ----------
    mode : str, optional
        Model mode: {"classification", "regression"}.
    verbose : int
        Verbosity for logger.
    """

    _add_to_settings = ["features_", *BaseFeatureProcessor._add_to_settings]

    def __init__(self, mode="notset", verbose=0):
        super().__init__(mode=mode, verbose=verbose)

        self.features_: List[str] = []
        self._validation_model = None
        self._baseline_scores = None

    def _fit_transform(self, x, y=None, **fit_params):
        raise NotImplementedError("Abstract method.")

    def _transform(self, x, y=None):
        raise NotImplementedError("Abstract method.")

    def set_features(self, features):
        """
        (Re-)set the features_ attribute.

        Parameters
        ----------
        features : typing.Iterable of str
        """
        # Check input
        if isinstance(features, str):
            features = [features]
        check_dtypes([("feature_item", x, str) for x in features])
        # Apply
        self.features_ = sorted(features)

    def add_features(self, features):
        """
        Add items to the features_ attribute.

        Parameters
        ----------
        features : typing.Iterable of str
        """
        # Check input
        if isinstance(features, str):
            features = [features]
        check_dtypes([("feature_item", x, str) for x in features])
        # Apply
        self.features_.extend(features)
        self.features_ = sorted(set(self.features_))

    def remove_features(self, features):
        """
        Remove items in the features_ attribute.

        Parameters
        ----------
        features : typing.Iterable of str
        """
        # Check input
        if isinstance(features, str):
            features = [features]
        check_dtypes([("feature_item", x, str) for x in features])
        # Check integrity
        if not set(features).issubset(self.features_):
            raise ValueError(
                f"Cannot remove features that are not existing: "
                f"{set(features) - set(self.features_)}"
            )
        # Apply
        self.features_ = [x for x in self.features_ if x not in features]

    def _set_validation_model(self):
        """
        Set the validation model for feature scoring.
        """
        assert self.mode in ("classification", "regression"), "Invalid mode."
        if self.mode == "classification":
            self._validation_model = DecisionTreeClassifier(
                max_depth=3,
                class_weight="balanced",
                random_state=19483,
            )
        else:
            self._validation_model = DecisionTreeRegressor(
                max_depth=3,
                random_state=19483,
            )

    def _calc_feature_scores(self, feature, y):
        """
        Analyses and scores a feature.

        Parameters
        ----------
        feature : pd.Series
            Feature to be analysed.
        y : pd.Series
            Target data (for scoring).

        Returns
        -------
        scores : array of float
            Feature score. In case of multiclass, a score per class.
        """
        # (Re-)fit validation model.
        #  Note that we do not make a train-test split. In this case, it makes sense as
        #  we only fit a shallow tree (max_depth=3). Because of that the model cannot
        #  really overfit.
        feature = feature.to_frame()
        self._validation_model.fit(feature, y)

        # Score
        if self.mode == "classification":
            classes = (
                self.classes_
                if hasattr(self, "classes_")
                else self._validation_model.classes_
            )
            if len(classes) > 2:
                scores = [
                    self._validation_model.score(feature, y, y == c)
                    if c in self._validation_model.classes_
                    else 0.0
                    for c in classes
                ]  # weighted score
            else:
                scores = [self._validation_model.score(feature, y)]  # average score

        elif self.mode == "regression":
            scores = [self._validation_model.score(feature, y)]  # average score

        else:
            raise AttributeError("Invalid mode.")

        return np.array(scores)

    def _init_feature_baseline_scores(self, x, y):
        """
        Initializes the baseline score of the given features.

        Parameters
        ----------
        x : pd.DataFrame
            Feature data.
        y : pd.Series
            Target data.
        """
        baseline_scores = x.apply(self._calc_feature_scores, y=y, axis=0)
        self._baseline_scores = baseline_scores.max(1).values.tolist()

        if self.mode == "classification":
            self.classes_ = sorted(y.unique())

    def _update_feature_baseline_scores(self, scores):
        """
        Update the baseline scores.

        Parameters
        ----------
        scores : pd.DataFrame
            Scores where each column contains the scores for the given feature.
        """
        scores = pd.concat([scores, pd.Series(self._baseline_scores)], axis=1)
        self._baseline_scores = scores.max(1).values.tolist()

    def accept_feature(self, scores):
        """
        Decides whether to accept a new feature.

        Parameters
        ----------
        scores : array of float
            Scores for checking against baseline threshold.

        Returns
        -------
        bool
            Whether to accept the feature.
        """
        if self._baseline_scores is None:
            warn("No baseline score is set. Output will be false", UserWarning)
        return any(scores >= 0.95 * np.array(self._baseline_scores))

    def select_scores(self, scores, best_n_per_class=50, update_baseline=False):
        """
        Scores and selects each feature column.

        Parameters
        ----------
        scores : pd.DataFrame
            Scores to be selected.
        best_n_per_class : int
            Limit for the number of accepted features per class (take the best).
        update_baseline : bool
            Whether to update the baseline scores.

        Returns
        -------
        pd.DataFrame
            Scores for accepted features.

        Notes
        -----
        For the scores dataframe, values represent the scores (per class) and column
        names the respective feature name.
        """
        check_dtypes([("scores", scores, pd.DataFrame)])

        if scores.shape[1] == 0:
            return scores

        # For each class, add take the best `best_n_per_class` features
        selection = set()
        for _, row in scores.iterrows():
            cls_selection = row.sort_values().iloc[-best_n_per_class:]
            selection = selection.union(cls_selection.index)
        selection = list(selection)

        # Accept features that are better than the baseline
        accept_mask = scores[selection].apply(self.accept_feature, axis=0)
        accepted_scores = scores[selection].loc[:, accept_mask]

        # Update baseline and return accepted feature names
        if update_baseline:
            self._update_feature_baseline_scores(accepted_scores)

        return accepted_scores



#  Copyright (c) 2022 by Amplo.

from copy import deepcopy

import numpy as np
import pandas as pd
import xgboost as xgb
from sklearn.model_selection import train_test_split


class XGBClassifier:
    _estimator_type = "classifier"
    default_params = {"verbosity": 0, "num_boost_round": 100}
    has_predict_proba = True

    def __init__(self, **params):
        """
        XG Boost wrapper
        @param params: Model parameters
        """
        self.num_boost_round = None
        self.params = None
        self.set_params(**params)
        self.classes_ = None
        self.model = None
        self.callbacks = None
        self.trained = False
        self.binary = False

    @staticmethod
    def convert_to_d_matrix(X, y=None):
        # Convert input
        assert type(X) in [
            pd.DataFrame,
            pd.Series,
            np.ndarray,
        ], "Unsupported data input format"
        if isinstance(X, np.ndarray) and len(X.shape) == 0:
            X = X.reshape((-1, 1))

        if y is None:
            return xgb.DMatrix(X)

        else:
            assert type(y) in [pd.Series, np.ndarray], "Unsupported data label format"
            return xgb.DMatrix(X, label=y)

    def fit(self, X, y):
        # Split & Convert data
        train_x, test_x, train_y, test_y = train_test_split(
            X, y, stratify=y, test_size=0.1
        )
        d_train = self.convert_to_d_matrix(train_x, train_y)
        d_test = self.convert_to_d_matrix(test_x, test_y)

        # Set loss function
        self.classes_ = np.unique(y)
        self.binary = len(self.classes_) == 2
        if self.binary:
            self.params["objective"] = "binary:logistic"
        else:
            self.params["objective"] = "multi:softprob"
            self.params["num_class"] = len(self.classes_)

        # Model training
        self.model = xgb.train(
            self.params,
            d_train,
            evals=[(d_test, "validation"), (d_train, "train")],
            verbose_eval=False,
            num_boost_round=self.num_boost_round,
            callbacks=[self.callbacks] if self.callbacks is not None else None,
            early_stopping_rounds=100,
        )
        self.trained = True

    def predict(self, X, *args, **kwargs):
        # todo check input data
        assert self.trained is True, "Model not yet trained"
        d_predict = self.convert_to_d_matrix(X)
        prediction = self.model.predict(d_predict, *args, **kwargs)

        # Parse into most-likely class
        if self.binary:
            return np.round(prediction).astype(int)
        else:
            return np.argmax(prediction, axis=1)

    def predict_proba(self, X, *args, **kwargs):
        # todo check input data
        assert self.trained is True, "Model not yet trained"
        d_predict = self.convert_to_d_matrix(X)
        prediction = self.model.predict(d_predict, *args, **kwargs)

        # Parse into probabilities
        if self.binary:
            return np.hstack((1 - prediction, prediction)).reshape((-1, 2), order="F")
        else:
            return prediction

    def set_params(self, **params):
        # Add default
        for k, v in self.default_params.items():
            if k not in params:
                params[k] = v

        # Remove fit args
        if "callbacks" in params:
            self.callbacks = params.pop("callbacks")

        # Update class
        self.num_boost_round = params.pop("num_boost_round")
        self.params = params
        return self

    def get_params(self, **args):
        params = deepcopy(self.params)
        if "deep" in args:
            return params
        if self.callbacks is not None:
            params["callbacks"] = self.callbacks
        params["num_boost_round"] = self.num_boost_round
        return params
